import { IFormSetting } from '.';
import { ICrossFieldConditions, IFormLayout, IFormOverrides, IFormV2 } from './models';
import { ICrossFieldValidations } from './modelsCommon';
import { IFormParsed, IFormSettings } from './modelsParsed';

export const parseFormV2 = (form: IFormV2): IFormParsed => {
  const {
    submitButtonText,
    cancelButtonText,
    validationObject,
    crossFieldValidations,
    crossFieldConditions,
    formLayout,
    overrides,
    children,
  } = form;

  const allFieldNames = Object.keys(validationObject.properties);

  const formOptions = allFieldNames.reduce((acc, fieldName) => {
    const fieldLayout = formLayout[fieldName] || {
      md: 12,
      sm: 12,
      xs: 12,
    };

    const formSetting: IFormSetting = { fieldLayout };

    if (crossFieldValidations?.[fieldName]) {
      formSetting.crossFieldValidations = crossFieldValidations[fieldName];
    }

    if (crossFieldConditions?.[fieldName]) {
      formSetting.crossFieldConditions = crossFieldConditions?.[fieldName];
    }

    if (overrides?.[fieldName]) {
      formSetting.overrides = overrides[fieldName];
    }

    acc[fieldName] = formSetting;

    return acc;
  }, {} as IFormSettings);

  const parsedForm: IFormParsed = {
    formatVersion: 'parsed',
    jsonSchema: validationObject,
    formSettings: formOptions,
    children,
    submitButtonText,
  };

  if (cancelButtonText) {
    parsedForm.cancelButtonText = cancelButtonText;
  }

  return parsedForm;
};

export const parseToFormV2 = (form: IFormParsed): IFormV2 => {
  const { submitButtonText, cancelButtonText, formSettings, jsonSchema, children } = form;

  const allFieldNames = Object.keys(formSettings);

  const crossFieldValidations: ICrossFieldValidations = {};

  const crossFieldConditions: ICrossFieldConditions = {};

  const formLayout: IFormLayout = {};

  const overrides: IFormOverrides = {};

  allFieldNames.forEach((key) => {
    const keyCrossFieldConditions = formSettings[key]?.crossFieldConditions;
    const keyCrossFieldValidations = formSettings[key]?.crossFieldValidations;
    const keyFieldOverrides = formSettings[key]?.overrides;

    if (keyCrossFieldConditions) {
      crossFieldConditions[key] = keyCrossFieldConditions;
    }

    if (keyCrossFieldValidations) {
      crossFieldValidations[key] = keyCrossFieldValidations;
    }

    if (keyFieldOverrides) {
      overrides[key] = keyFieldOverrides;
    }

    formLayout[key] = formSettings[key]?.fieldLayout || {
      md: 12,
      sm: 12,
      xs: 12,
    };
  });

  const formV2: IFormV2 = {
    formatVersion: 2,
    crossFieldConditions,
    crossFieldValidations,
    formLayout,
    overrides,
    validationObject: jsonSchema,
    children,
    submitButtonText,
  };

  if (cancelButtonText) {
    formV2.cancelButtonText = cancelButtonText;
  }

  return formV2;
};
